use clap::{Arg, Command};
use log::LevelFilter;
use scootaloo::*;
use simple_logger::SimpleLogger;
use std::str::FromStr;

const DEFAULT_CONFIG_PATH: &str = "/usr/local/etc/scootaloo.toml";

fn main() {
    let matches = Command::new(env!("CARGO_PKG_NAME"))
        .version(env!("CARGO_PKG_VERSION"))
        .about("A Twitter to Mastodon bot")
        .arg(
            Arg::new("config")
                .short('c')
                .long("config")
                .value_name("CONFIG_FILE")
                .help(&format!(
                    "TOML config file for scootaloo (default {})",
                    DEFAULT_CONFIG_PATH
                ))
                .num_args(1)
                .default_value(DEFAULT_CONFIG_PATH)
                .display_order(1),
        )
        .arg(
            Arg::new("log_level")
                .short('l')
                .long("loglevel")
                .value_name("LOGLEVEL")
                .help("Log level. Valid values are: Off, Warn, Error, Info, Debug")
                .num_args(1)
                .value_parser(["Off", "Warn", "Error", "Info", "Debug"])
                .display_order(2),
        )
        .subcommand(
            Command::new("register")
                .version(env!("CARGO_PKG_VERSION"))
                .about("Command to register to a Mastodon Instance")
                .arg(
                    Arg::new("host")
                        .short('H')
                        .long("host")
                        .value_name("HOST")
                        .help("Base URL of the Mastodon instance to register to (no default)")
                        .num_args(1)
                        .required(true)
                        .display_order(1)
                    )
                .arg(
                    Arg::new("name")
                        .short('n')
                        .long("name")
                        .help("Twitter Screen Name (like https://twitter.com/screen_name, no default)")
                        .num_args(1)
                        .required(true)
                        .display_order(2)
                ),
        )
        .subcommand(
            Command::new("init")
                .version(env!("CARGO_PKG_VERSION"))
                .about("Command to init Scootaloo DB")
                .arg(
                    Arg::new("config")
                        .short('c')
                        .long("config")
                        .value_name("CONFIG_FILE")
                        .help(&format!(
                            "TOML config file for scootaloo (default {})",
                            DEFAULT_CONFIG_PATH
                        ))
                        .default_value(DEFAULT_CONFIG_PATH)
                        .num_args(1)
                        .display_order(1),
                ),
        )
        .subcommand(
            Command::new("migrate")
            .version(env!("CARGO_PKG_VERSION"))
            .about("Command to migrate Scootaloo DB")
            .arg(
                Arg::new("config")
                    .short('c')
                    .long("config")
                    .value_name("CONFIG_FILE")
                    .help(&format!("TOML config file for scootaloo (default {})", DEFAULT_CONFIG_PATH))
                    .default_value(DEFAULT_CONFIG_PATH)
                    .num_args(1)
                    .display_order(1),
            )
            .arg(
                    Arg::new("name")
                        .short('n')
                        .long("name")
                        .help("Twitter Screen Name (like https://twitter.com/screen_name, no default)")
                        .num_args(1)
                        .display_order(2)
            )
        )
        .get_matches();

    match matches.subcommand() {
        Some(("register", sub_m)) => {
            register(
                sub_m.get_one::<String>("host").unwrap(),
                sub_m.get_one::<String>("name").unwrap(),
            );
            return;
        }
        Some(("init", sub_m)) => {
            let config = parse_toml(sub_m.get_one::<String>("config").unwrap());
            init_db(&config.scootaloo.db_path).unwrap();
            return;
        }
        Some(("migrate", sub_m)) => {
            let config = parse_toml(sub_m.get_one::<String>("config").unwrap());
            let config_twitter_screen_name =
                &config.mastodon.values().next().unwrap().twitter_screen_name;
            migrate_db(
                &config.scootaloo.db_path,
                sub_m
                    .get_one::<String>("name")
                    .unwrap_or(config_twitter_screen_name),
            )
            .unwrap();
            return;
        }
        _ => (),
    }

    if let Some(level) = matches.get_one::<String>("log_level") {
        SimpleLogger::new()
            .with_level(LevelFilter::from_str(level).unwrap())
            .init()
            .unwrap();
    }

    let config = parse_toml(matches.get_one::<String>("config").unwrap());

    run(config);
}
